use rusqlite::OptionalExtension;
use tokio::sync::{
	mpsc::{unbounded_channel, UnboundedReceiver, UnboundedSender},
	oneshot,
};

pub struct DbExecutor {
	rx: UnboundedReceiver<Task>,
	db: rusqlite::Connection,
}

impl DbExecutor {
	pub fn create(dbpath: &str) -> rusqlite::Result<(Self, ExecutorConnection)> {
		let (tx, rx) = unbounded_channel();
		let db = rusqlite::Connection::open(dbpath)?;
		db.execute(
			"create table if not exists pastes( \
            id integer primary key, paste text not null, \
            author text not null, \
            timestamp integer default (strftime('%s','now')))",
			[],
		)?;
		Ok((Self { rx, db }, ExecutorConnection { tx }))
	}

	pub fn run(&mut self) {
		while let Some(task) = self.rx.blocking_recv() {
			match task {
				Task::GetPaste(tx, v) => {
					let result = self
						.db
						.query_row("select paste from pastes where id=?", [v], |r| r.get(0))
						.optional();
					let _ = tx.send(result);
				}
				Task::AddPaste(tx, v, a) => {
					let result = self.db.query_row::<i64, _, _>(
						"insert into pastes(paste,author) values(?,?) returning id",
						[v, a],
						|r| r.get(0),
					);
					let _ = tx.send(result);
				}
				Task::Cleanup(tx, mins) => {
					let result = self.db.execute(
						"delete from pastes where timestamp < strftime('%s','now')-?",
						[mins],
					);
					let _ = tx.send(result);
				}
			}
		}
	}
}

pub struct ExecutorConnection {
	tx: UnboundedSender<Task>,
}

impl Clone for ExecutorConnection {
	fn clone(&self) -> Self {
		Self {
			tx: self.tx.clone(),
		}
	}
}

macro_rules! executor_wrapper {
    ($name:ident, $task:expr, $ret:ty, $($arg:ident: $ty:ty),*) => {
        pub async fn $name(&self, $($arg: $ty),*) -> $ret {
          let (otx, orx) = oneshot::channel();
          self.tx.send($task(otx, $($arg),*)).unwrap();
          orx.await.unwrap()
        }
    };
}

#[derive(Debug)]
enum Task {
	GetPaste(oneshot::Sender<rusqlite::Result<Option<String>>>, i64),
	AddPaste(oneshot::Sender<rusqlite::Result<i64>>, String, String),
	Cleanup(oneshot::Sender<rusqlite::Result<usize>>, i64),
}

impl ExecutorConnection {
	executor_wrapper!(
		add_paste,
		Task::AddPaste,
		rusqlite::Result<i64>,
		paste: String,
		author: String
	);
	executor_wrapper!(
		get_paste,
		Task::GetPaste,
		rusqlite::Result<Option<String>>,
		id: i64
	);
	executor_wrapper!(cleanup, Task::Cleanup, rusqlite::Result<usize>, retention: i64);
}
