use std::{env, fs, net::SocketAddr, str::FromStr, sync::Arc, thread};

use base64::Engine;
use tokio::{
	io::{AsyncReadExt, AsyncWriteExt},
	net::{TcpListener, TcpStream},
	select,
	sync::{
		broadcast,
		mpsc::{self, Sender},
	},
	time::{sleep, Duration},
};
use tracing::Level;

use database::{DbExecutor, ExecutorConnection};

mod cfg;
mod database;

const BUFSIZE: u64 = u16::MAX as u64 + 1;
const B64: base64::engine::GeneralPurpose = base64::engine::general_purpose::STANDARD_NO_PAD;

#[cfg(unix)]
async fn terminate_signal() {
	use tokio::signal::unix::{signal, SignalKind};
	let mut sigterm = signal(SignalKind::terminate()).unwrap();
	let mut sigint = signal(SignalKind::interrupt()).unwrap();
	select! {
		_ = sigterm.recv() => {},
		_ = sigint.recv() => {}
	}
}

#[cfg(windows)]
async fn terminate_signal() {
	use tokio::signal::windows::ctrl_c;
	let mut ctrlc = ctrl_c().unwrap();
	let _ = ctrlc.recv().await;
}

#[tokio::main(flavor = "current_thread")]
async fn main() -> anyhow::Result<()> {
	let config_var = env::var("QBIN_CONFIG");
	let config_path = config_var.as_deref().unwrap_or("qbin.toml");
	println!("Loading config from '{}'...", config_path);
	let config_str = fs::read_to_string(config_path)?;
	let config: Arc<cfg::Config> = Arc::new(toml::from_str(&config_str)?);

	tracing_subscriber::fmt::fmt()
		.with_max_level(Level::from_str(&config.log)?)
		.init();

	tracing::info!(concat!("Initializing - qbin ", env!("CARGO_PKG_VERSION")));

	let (mut exec, execonn) = DbExecutor::create(&config.database)?;
	let executor_thread = thread::spawn(move || {
		exec.run();
		tracing::info!("Database executor has been shut down");
	});

	let listener = TcpListener::bind(&config.listen).await?;
	tracing::info!(
		"Socket bound to {}! Listening for connections...",
		&config.listen
	);
	let (cancel_send, _cancel_token) = broadcast::channel::<()>(1);
	let (cancelled_send, mut cancelled_recv) = mpsc::channel::<()>(1);
	let cleanup = tokio::spawn(cleanup_task(
		execonn.clone(),
		config.retention,
		config.interval,
		cancel_send.subscribe(),
		cancelled_send.clone(),
	));

	loop {
		select! {
			Ok((socket, remote)) = listener.accept() => {
				tracing::debug!("New connection from {}", remote);
				tokio::spawn(execute_client(execonn.clone(), socket, remote, cancel_send.subscribe(), cancelled_send.clone()));
			},
			_ = terminate_signal() => break
		}
	}

	tracing::info!("Shutting down...");
	cancel_send.send(()).unwrap();
	drop(cancelled_send);
	let _ = cancelled_recv.recv().await;
	cleanup.await.unwrap();
	drop(execonn);
	executor_thread.join().unwrap();
	Ok(())
}

async fn cleanup_task(
	db: ExecutorConnection,
	retention: i64,
	interval: i64,
	mut cancel: broadcast::Receiver<()>,
	_cancel_handle: Sender<()>,
) {
	loop {
		tracing::debug!("Starting cleanup...");
		match db.cleanup(retention).await {
			Err(e) => {
				tracing::error!("Database cleanup has failed: {}", e);
			}
			Ok(o) => {
				tracing::debug!("Database cleanup succeeded, affected {} rows", o);
			}
		}
		select! {
			_ = sleep(Duration::from_secs(interval as u64)) => {},
			_ = cancel.recv() => {
				break
			}
		}
	}
	tracing::info!("Cleanup task has been shut down");
}

async fn handle_client(
	db: ExecutorConnection,
	mut sock: TcpStream,
	remote: SocketAddr,
) -> anyhow::Result<()> {
	let mut inputstring = String::new();
	let (sockread, mut sockwrite) = sock.split();
	let bytes = sockread
		.take(BUFSIZE)
		.read_to_string(&mut inputstring)
		.await? as u64;

	if bytes == 0 {
		return Ok(());
	}
	if bytes == BUFSIZE {
		tracing::debug!("User {} exceeded the input size limit", remote);
		sockwrite
			.write_all(b"[!] Your input was too large.\n")
			.await?;
		return Ok(());
	}

	let trimmed = inputstring.trim();
	if trimmed.len() > 2 && trimmed.starts_with("G!") {
		let pasteid_raw = &trimmed[2..];
		let pasteid = match B64
			.decode(pasteid_raw)
			.ok()
			.and_then(|v| <[u8; 8]>::try_from(v).ok())
			.and_then(|v| Some(i64::from_le_bytes(v)))
		{
			None => {
				tracing::debug!(
					"User {} has entered an invalid paste ID: {:?}",
					remote,
					pasteid_raw
				);
				sockwrite
					.write_all(b"[!] Invalid paste ID format.\n")
					.await?;
				return Ok(());
			}
			Some(o) => o,
		};
		match db.get_paste(pasteid).await? {
			Some(o) => {
				tracing::debug!("User {} has opened the paste '{}'", remote, pasteid_raw);
				sockwrite.write_all(o.as_bytes()).await?;
			}
			None => {
				tracing::debug!(
					"User {} tried to retreive an invalid paste: {}",
					remote,
					pasteid_raw
				);
				sockwrite.write_all(b"[!] Paste not found\n").await?;
			}
		}
	} else {
		let paste_id = db.add_paste(inputstring, remote.ip().to_string()).await?;
		let encoded_id = B64.encode(paste_id.to_le_bytes());
		tracing::debug!("Paste '{}' has been created by {}.", encoded_id, remote);
		sockwrite
			.write_all(format!("{}\n", encoded_id).as_bytes())
			.await?;
	}
	Ok(())
}

async fn execute_client(
	db: ExecutorConnection,
	sock: TcpStream,
	remote: SocketAddr,
	mut cancel: broadcast::Receiver<()>,
	_cancel_handle: Sender<()>,
) {
	select! {
		v = handle_client(db, sock, remote) => {
			match v {
				Ok(_) => tracing::debug!("Connection with {} closed", remote),
				Err(e) => tracing::warn!("Error on connection with {}: {}", remote, e)
			}
		},
		_ = cancel.recv() => {
			tracing::debug!("Connection with {} closed due to global shutdown", remote);
			return
		}
	}
}
